#!/usr/bin/env python
"""flashtorch.utils

This module provides utility functions for image handling and tensor
transformation.

"""
from PIL import Image

import matplotlib.pyplot as plt

import torchvision.transforms as transforms

from .imagenet import *

def load_image(image_path):
    """Loads image as a PIL RGB image.

    Args:
        image_path (str): A path to the image

    Returns:
        An instance of PIL.Image.Image in RGB

    """

    return Image.open(image_path).convert('RGB')


def apply_transforms(image, size=224):
    """Transforms a PIL image to torch.Tensor.

    Applies a series of tranformations on PIL image including a conversion
    to a tensor. The returned tensor has a shape of :math:`(N, C, H, W)` and
    is ready to be used as an input to neural networks.

    First the image is resized to 256, then cropped to 224. The `means` and
    `stds` for normalisation are taken from numbers used in ImageNet, as
    currently developing the package for visualizing pre-trained models.

    The plan is to to expand this to handle custom size/mean/std.

    Args:
        image (PIL.Image.Image): RGB PIL Image
        size(int, optional, default=224): Desired size (width/height) of the
            output tensor

    Shape:
        Input: N/A
        Output: :math:`(N, C, H, W)`

    Returns:
        torch.Tensor (torch.float32): Transformed image tensor

    Note:
        Symbols used to describe dimensions:
            - N: number of images in a batch
            - C: number of channels
            - H: height of the image
            - W: width of the image

    """

    means = [0.485, 0.456, 0.406]
    stds = [0.229, 0.224, 0.225]

    transform = transforms.Compose([
        transforms.Resize(size),
        transforms.CenterCrop(size),
        transforms.ToTensor(),
        transforms.Normalize(means, stds)
    ])

    return transform(image).unsqueeze(0)


def denormalize(tensor):
    """Reverses the normalisation on a tensor.

    Performs a reverse operation on a tensor, so the pixel value range is
    between 0 and 1. Useful for when plotting a tensor into an image.

    Normalisation: (image - mean) / std
    Denormalisation: image * std + mean

    Args:
        tensor (torch.Tensor, dtype=torch.float32): Normalized image tensor

    Shape:
        Input: :math:`(N, C, H, W)`
        Output: :math:`(N, C, H, W)` (same shape as input)

    Return:
        torch.Tensor (torch.float32): Demornalised image tensor with pixel
            values between [0, 1]

    Note:
        Symbols used to describe dimensions:
            - N: number of images in a batch
            - C: number of channels
            - H: height of the image
            - W: width of the image

    """

    means = [0.485, 0.456, 0.406]
    stds = [0.229, 0.224, 0.225]

    denormalized = tensor.clone()

    for channel, mean, std in zip(denormalized[0], means, stds):
        channel.mul_(std).add_(mean)

    return denormalized


def normalize(tensor, min_value=0.0, max_value=1.0):
    """Normalises input tensor values between min/max (default: 0.0/1.0).

    Args:
        tensor (torch.Tensor)
        min_value (float, optional, default=0.0)
        max_value (float, optional, default=1.0)

    Return:
        torch.Tensor (torch.float32): Demornalised tensor with values between
            [min_value, max_value]

    """

    target_range = max_value - min_value

    tensor_min = tensor.min()
    tensor_max = tensor.max()
    tensor_range = tensor_max - tensor_min

    eps = 1e-7

    # Normalise to [0, 1] (using epsilon to avoid diving by zero)

    normalized = (tensor - tensor_min) / (tensor_max - tensor_min + eps)

    # Scale it with min/max_value

    return (max_value - min_value) * normalized + min_value


def format_for_plotting(tensor):
    """Formats the shape of tensor for plotting.

    Tensors typically have a shape of :math:`(N, C, H, W)` or :math:`(C, H, W)`
    which is not suitable for plotting as images. This function formats an
    input tensor :math:`(H, W, C)` for RGB and :math:`(H, W)` for mono-channel
    data.

    Args:
        tensor (torch.Tensor, torch.float32): Image tensor

    Shape:
        Input: :math:`(N, C, H, W)` or :math:`(C, H, W)`
        Output: :math:`(H, W, C)` or :math:`(H, W)`, respectively

    Return:
        torch.Tensor (torch.float32): Formatted image tensor (detached)

    Note:
        Symbols used to describe dimensions:
            - N: number of images in a batch
            - C: number of channels
            - H: height of the image
            - W: width of the image

    """

    has_batch_dimension = len(tensor.shape) == 4
    formatted = tensor.clone()

    if has_batch_dimension:
        formatted = tensor.squeeze(0)

    if formatted.shape[0] == 1:
        return formatted.squeeze(0).detach()
    else:
        return formatted.permute(1, 2, 0).detach()


def visualize(input_, gradients, max_gradients, cmap='seismic', alpha=0.5):

    input_ = format_for_plotting(denormalize(input_))
    gradients = format_for_plotting(normalize(gradients))
    max_gradients = format_for_plotting(normalize(max_gradients))

    subplots = [
        # (title, [(image1, cmap, alpha), (image2, cmap, alpha)])
        ('Input image', [(input_, None, None)]),
        ('Gradients across RGB channels', [(gradients, None, None)]),
        ('Max gradients', [(max_gradients, cmap, None)]),
        ('Overlay', [(input_, 'gray', None), (max_gradients, cmap, alpha)])
    ]

    num_subplots = len(subplots)

    fig = plt.figure(figsize=(16, 4))


    for i, (title, images) in enumerate(subplots):
        ax = fig.add_subplot(1, num_subplots, i + 1)
        ax.set_axis_off()

        for image, cmap, alpha in images:
            ax.imshow(image, cmap=cmap, alpha=alpha)

        ax.set_title(title)
